%
% the ECLiPSe solver module
%
% This is the environment where a compiled CLP will execute.
%

:- use_module(library(fd)).		%use the FD library
:- use_module(library(propia)).		%use the Propia library
%:- ['../../prolog/util/is_daplex_variable'].
:- ['../../prolog/util/retractall'].
:- [eclipse_stdlib].



:- dynamic(solve/1).			%the top-level predicate
:- dynamic(declare_vars/1).
:- dynamic(post_constraints/1).		%clauses posting constraints on vars
:- dynamic(post_constraint/2).		%clauses posting constraints on vars

:- dynamic(object/2).			%objects
:- dynamic(fnval/5).			%relationships
:- dynamic(cast/3).			%casting information
%:- [code2].
%
% the 2 main interfacing predicates
%

tell_solver(Something) :-
	assert(Something).

ask_solver(Predicate) :-
	call(Predicate).

clean_up :-
	retractall(solve(_)),
	retractall(declare_vars(_)),
	retractall(post_constraints(_)),
	retractall(post_constraint(_,_)),
	retractall(object(_,_)),
	retractall(fnval(_,_,_,_,_)),
	retractall(cast(_,_,_)),
	!.

clean_up.

add_tuples([]).
add_tuples([T|Ts]) :-
	add_tuples(T),
	add_tuples(Ts).

add_tuples(T) :-
	convert_term_arg_into_atom(T,A),
	(	(	call(A),
			!
		);
		assert(A)
	).

%
% convert term in arguments into atoms
%

%
%cater for object/2
%
convert_term_arg_into_atom(object(Class,Var),object(Class,Var_as_atom)) :-
	!,
	convert_term_into_atom([Var],[Var_as_atom]).

%
%cater for fnval/5
%
convert_term_arg_into_atom(T,A) :-
	T=fnval(Fname,Types,Vars,Result_type,Result),
	!,
	convert_term_into_atom([Result|Vars],[Result_as_atom|Vars_as_atoms]),
	A=fnval(Fname,Types,Vars_as_atoms,Result_type,Result_as_atom).

%
%cater for cast/3
%
convert_term_arg_into_atom(T,A) :-
	T=cast(Class,From,To),
	!,
	convert_term_into_atom([From,To],[From_as_atom,To_as_atom]),
	A=cast(Class,From_as_atom,To_as_atom).

convert_term_arg_into_atom(T,T).

convert_term_into_atom([],[]).

convert_term_into_atom([T|Ts],[T|As]) :-
	integer(T),
	!,
	convert_term_into_atom(Ts,As).

convert_term_into_atom([T|Ts],[T|As]) :-
	atom(T),
	!,
	convert_term_into_atom(Ts,As).

convert_term_into_atom([T|Ts],[A|As]) :-
	!,
	term_string(T,S),
	atom_string(A,S),
	convert_term_into_atom(Ts,As).

/*
%
% casting constraint betwen 2 objects & a class name
%
cast(Target_class,From,From) :-
	object(Target_class,From).

cast(Target_class,From,To) :-
	cast(Target_class,To,_,From).

cast(Target_class,From,To) :-
	cast(_,From,Target_class,To).
*/

