%
% ECLiPSe wrapper
%

:- ['../../prolog/util/ksl'].				%the KRAFT service layer
:- ['../../prolog/messaging/messaging'].		%the messaging module
:- ['../../prolog/util/cmdline'].			%the command-line module
:- ['../../prolog/util/is_daplex_variable'].
:- ['../../prolog/util/cif2i'].			%util that converts CIF to ICode
:- [eclipse_solver].					%load solver environment for KRAFTA

%:- use_module(library(iso)).
%:- use_module(library(lists)).
%:- [domain].
%:- ['../../prolog/util/cif_util'].			%general CIF manipulating predicates

%
% different starting predicates
%
% go		:	Read config file name from command-line arg "-c filename"
%			If no "-c" is given, use the default "ew.config".
%			Optional linda host & port given by "-lh" & "-lp".
%
% go(Host,Port)	:	Use the linda server at Host:Port.
%

go :-
	ksl:read_agent_config,
	ksl:get_config(linda_host,Host),
	ksl:get_config(linda_port,Port),
	init(Host,Port),
	main_loop.


%
% initializing the agent
%
init(Host,Port) :-
	write('*** initialising ECLiPSe Wrapper...'),
	nl,
	messaging:init_messaging([server_host:Host,server_port:Port]),
	local_init.

%
% shutting the agent
%
shutdown :-
	write('*** shutting down ECLiPSe Wrapper...'),
	nl,
	local_shutdown,
	messaging:halt_messaging.

%
% local init & shutdown
%
local_init :-
	clean_up.

local_shutdown.

%
% the main loop
%
main_loop :-
	write('*** starting ECLiPSe Wrapper main loop'),
	nl,
	ksl:get_config(name,Me),
	repeat,
	write('... waiting for incoming message...'),
	nl,
	receive_incoming(Me,Message),
	write('... something received...'),
	nl,
	messaging:decompose_message(Sender,Me,Perf,Param,Message),
	write('... handling income message...'),
	nl,
	handle_message(Sender,Me,Perf,Param,Message).		%MUST fail for looping

receive_incoming(Me,Message) :-
	messaging:receive_message(Me,Message),
	!.

%
% handling different messages
%

%
%PING
%
handle_message(Sender,Me,ping,Params,Message) :-
	write('>>> PING received'),
	nl,
	messaging:decompose_message(Context,_,Message),
	messaging:decompose_context(Version,Cid,_,_,_,Context),
	Check_list=[if_not(reply_with:_,'reply_with not found')],
	check_params(Params,Check_list,_,Error),
	(	(	%
			% no problem
			%
			Error=none,
			!,
			messaging:get_argument(Params,reply_with:Reply_with),		%get reply_with
			messaging:set_argument(in_reply_to:Reply_with,Reply_params),	%use as in_reply_to
			messaging:compose_ccql_core(pong,Reply_params,Reply_core),	%compose PONG core
			write('... PONG composed'),
			nl
		);
		(	messaging:set_argument(content:Error,Reply_params),
			messaging:compose_ccql_core(error,Reply_params,Reply_core),	%compose ERROR core
			write('... ERROR composed'),
			nl
		)

	),
	messaging:compose_context(Version,Cid,Me,Sender,_,Reply_context),	%compose the context
	messaging:compose_message(Reply_context,Reply_core,Reply),		%compose message
	write('<<< sending reply'),
	nl,
	messaging:send_message(Reply),
%	sleep(1),
	!,
	fail.


%
% ASK_IF
%
handle_message(Sender,Me,ask_if,Params,Message) :-
	write('>>> ASK_IF received'),
	nl,
	Check_list=[	if_not(language:eclipse,'language must be eclipse'),
			if_not(ontology:shared,'ontology must be shared'),
			if_not(reply_with:_,'reply_with not found'),
			if_not(content:_,'content missing')
			],
	check_params(Params,Check_list,_,Error),
	(	(	%
			% no error
			% and I can handle this ask_if
			%
			Error=none,
			messaging:get_argument(Params,[	reply_with:Reply_with,		%get reply_with
							content:Content]),		%and content
			execute_query(ask_if,Content,Result),
			!,
			messaging:set_argument([	in_reply_to:Reply_with,
							content:Result
						],
							Reply_params),	%use as in_reply_to
			messaging:compose_ccql_core(tell,Reply_params,Reply_core),	%compose TELL core
			write('... TELL composed'),
			nl

		);
		(	%
			% no error, but I can't handle this ask_if
			%
			Error=none,
			!,
			messaging:get_argument(Params,[	reply_with:Reply_with,		%get reply_with
							content:Pattern]),		%and pattern
			messaging:set_argument([	in_reply_to:Reply_with,
							content:Pattern
							],
							Reply_params),	%use as in_reply_to
			messaging:compose_ccql_core(sorry,Reply_params,Reply_core),	%compose TELL core
			write('... SORRY composed'),
			nl
		);
		(	messaging:set_argument(content:Error,Reply_params),
			messaging:compose_ccql_core(error,Reply_params,Reply_core),	%compose error core
			write('... ERROR composed'),
			nl
		)
	),
	messaging:decompose_message(Context,_,Message),
	messaging:decompose_context(Version,Cid,_,_,_,Context),
	messaging:compose_context(Version,Cid,Me,Sender,_,Reply_context),	%compose the context
	messaging:compose_message(Reply_context,Reply_core,Reply),		%compose message
	write('<<< sending reply'),
	nl,
	messaging:send_message(Reply),
%	sleep(1),
	!,
	fail.

%
% ASK_ONE
%
handle_message(Sender,Me,ask_one,Params,Message) :-
	write('>>> ASK_ONE received'),
	nl,
	Check_list=[	if_not(language:eclipse,'language must be eclipse'),
			if_not(ontology:shared,'ontology must be shared'),
			if_not(reply_with:_,'reply_with not found'),
			if_not(content:_,'content missing')
			],
	check_params(Params,Check_list,_,Error),				%check for any error
	handle_ask_one(Sender,Me,Params,Error,Reply_core),			%handle ASK_ONE
	messaging:decompose_message(Context,_,Message),				%get original context
	messaging:decompose_context(Version,Cid,_,_,_,Context),			%get versio & cid
	messaging:compose_context(Version,Cid,Me,Sender,_,Reply_context),	%compose the context
	messaging:compose_message(Reply_context,Reply_core,Reply),		%compose message
	write('<<< sending reply'),
	nl,
	messaging:send_message(Reply),
%	sleep(1),
	!,
	fail.



%
% ASK_ALL
%
handle_message(Sender,Me,ask_all,Params,Message) :-
	write('>>> ASK_ALL received'),
	nl,
	Check_list=[	if_not(language:eclipse,'language must be eclipse'),
			if_not(ontology:shared,'ontology must be shared'),
			if_not(reply_with:_,'reply_with not found'),
			if_not(content:_,'content missing')
			],
	check_params(Params,Check_list,_,Error),				%check for any error
	handle_ask_all(Sender,Me,Params,Error,Reply_core),			%handle ASK_ONE
	messaging:decompose_message(Context,_,Message),				%get original context
	messaging:decompose_context(Version,Cid,_,_,_,Context),			%get versio & cid
	messaging:compose_context(Version,Cid,Me,Sender,_,Reply_context),	%compose the context
	messaging:compose_message(Reply_context,Reply_core,Reply),		%compose message
	write('<<< sending reply'),
	nl,
	messaging:send_message(Reply),
%	sleep(1),
	!,
	fail.

%
% TELL
%
handle_message(Sender,Me,tell,Params,_Message) :-
	write('>>> TELL received'),
	nl,
	Check_list=[	if_not(language:eclipse,'language must be eclipse'),
			if_not(ontology:shared,'ontology must be shared'),
%			if_not(reply_with:_,'reply_with not found'),
			if_not(content:_,'content missing')
			],
	check_params(Params,Check_list,_,Error),				%check for any error
	handle_tell(Sender,Me,Params,Error,_Reply_core),			%handle TELL
	!,
	fail.


%
% HALT
%
handle_message(_,_,halt,_,_) :-
	write('W>>> agent terminating...'),
	nl,
	!.

%
% DEBUG
%
handle_message(_,_,debug,_,_) :-
	write('DEBUGGING invoked by message...'),
	nl,
	send_register,
	send_advertise,
	!,
	fail.

%
% ALL OTHERS
%
handle_message(Sender,Me,Perf,Params,Message) :-
	write('*** WARNING: unknown performative received'),
	nl,
	write('... composing SORRY'),
	nl,
	%
	% compose SORRY to sender
	%
	name(Perf,Perf_string),
	flatten(["cannot handle ",Perf_string],Error_string),			%compose error message string
	name(Error_message,Error_string),
	Check_list=[if_not(reply_with:_,'reply_with not found')],		%check for params
	check_params(Params,Check_list,_,Error),
	(	(	Error=none,						%if no problem w/ reply-with
			!,
			messaging:get_argument(Params,[reply_with:Reply_with]),	%get reply-with
			messaging:set_argument([in_reply_to:Reply_with,		%compose reply params w/ in-reply-to & content
						content:Error_message],
						Reply_params)
			);
		messaging:set_argument([content:Error_message],Reply_params)	%otherwise just put in the content
	),
	messaging:compose_ccql_core(sorry,Reply_params,Reply_core),		%compose SORRY core
	messaging:decompose_message(Context,_,Message),				%get original context
	messaging:decompose_context(Version,Cid,_,_,_,Context),			%get version & cid
	messaging:compose_context(Version,Cid,Me,Sender,_,Reply_context),	%compose the context
	messaging:compose_message(Reply_context,Reply_core,Reply),		%compose message
	messaging:send_message(Reply),
	write('<<< SORRY sent'),
	nl,
%	flush_output(user),
	flush(user),
	!,
	fail.


%
% check if params satisfy the required value
% return error message if appropriate
%
check_params(_,[],_:_,none).

check_params(Param,[C|Cs],Problematic_pair,Error) :-
	C=if_not(Arg:Expected_value,This_error),
	(	(	var(Expected_value),					%expecting any value
			messaging:get_argument(Param,Arg:Expected_value),	%found it there
			!,
			check_params(Param,Cs,Problematic_pair,Error)		%proceed
		);
		(	nonvar(Expected_value),					%expecting a certain value
			messaging:get_argument(Param,Arg:Current_value),	%get current one
			nonvar(Current_value),					%which is not a var
			Current_value=Expected_value,				%and they can unify
			!,
			check_params(Param,Cs,Problematic_pair,Error)
		);
		(	nonvar(Expected_value),
			messaging:get_argument(Param,Arg:Current_value),
			var(Current_value),
			!,
			Error=This_error,
			Problematic_pair=Arg:Current_value
		);
		(	Error=This_error,
			Problematic_pair=Arg:Expected_value
		)
	).


%
% handling different situations of ASK_ONE
%
%
handle_ask_one(Sender,Me,Params,none,Reply_core) :-
	%
	% no error, and I can handle it
	%
	messaging:get_argument(Params,[	reply_with:Reply_with,		%get reply_with
					content:Query]),		%get query as a goal
	execute_query(ask_one,Query,Result),
	messaging:set_argument([sender:Me,				%reply back
				receiver:Sender,
				in_reply_to:Reply_with,
				content:Result
				],
				Reply_params),				%use as in_reply_to
	messaging:compose_ccql_core(tell,Reply_params,Reply_core).	%compose TELL core


handle_ask_one(Sender,Me,Params,Error,Reply_core) :-
	%
	% has error
	%
	(	(	messaging:get_argument(Params,[reply_with:Reply_with]),
			!,
			messaging:set_argument([sender:Me,
						receiver:Sender,
						in_reply_with:Reply_with,
						language_cif,
						ontology:shared,
						content:Error
						],Reply_params)
		);
		messaging:set_argument([sender:Me,
					receiver:Sender,
					in_reply_with:Reply_with,
					language_cif,
					ontology:shared,
					content:Error
					],Reply_params)
	),
	messaging:compose_ccql_core(error,Reply_params,Reply_core),	%compose error core
	write('... ERROR composed'),
	nl.


%
% handling different situations of ASK_ALL
%
handle_ask_all(Sender,Me,Params,none,Reply_core) :-
	%
	% no error, and I can handle it
	%
	messaging:get_argument(Params,[	reply_with:Reply_with,		%get reply_with
					content:Query]),		%get query as a goal
	execute_query(ask_all,Query,Result),
	messaging:set_argument([sender:Me,
				receiver:Sender,
				in_reply_to:Reply_with,
				content:Result
				],
				Reply_params),				%use as in_reply_to
	messaging:compose_ccql_core(tell,Reply_params,Reply_core).	%compose TELL core


handle_ask_all(Sender,Me,Params,Error,Reply_core) :-
	%
	% has error
	%
	(	(	messaging:get_argument(Params,[reply_with:Reply_with]),
			!,
			messaging:set_argument([sender:Me,
						receiver:Sender,
						in_reply_with:Reply_with,
						language_cif,
						ontology:shared,
						content:Error
						],Reply_params)
		);
		messaging:set_argument([sender:Me,
					receiver:Sender,
					in_reply_with:Reply_with,
					language_cif,
					ontology:shared,
					content:Error
					],Reply_params)
	),
	messaging:compose_ccql_core(error,Reply_params,Reply_core),	%compose error core
	write('... ERROR composed'),
	nl.


%
% handle TELL
%
handle_tell(_Sender,_Me,Params,_Error,_Reply_core) :-
	messaging:get_argument(Params,content:Something),		%get content to tell
	tell_solver(Something).

%
% handle ASK_* by executing goal
%

execute_query(ask_if,Content,true) :-
	call(Content),
	!,
	write(Content),
	nl.

execute_query(ask_one,Content,Content) :-
	call(Content),
	!,
	write(Content),
	nl.

execute_query(ask_all,Content,Result) :-
	findall(Content,call(Content),Result),
	!,
	write(Result),
	nl.

execute_query(_,_,fail).

%
% starts after loading
%

%:- go.

