%
% compiling Daplex schema into RDFS
%
% By: K Hui
% Date: 12-4-2002
%

:- ['../../prolog/pillow/pillow'].		%load pillow
:- ['../../prolog/util/print_xml'].
:- ['../../prolog/util/concat_atoms'].

%:- daplex('cif.daplex').			%load Daplex schema of CIF

%
% known URLs
%
url(rdfs,'http://www.w3.org/2000/01/rdf-schema').
url(rdf,'http://www.w3.org/1999/02/22-rdf-syntax-ns').


go :-
	Xml=[Xml_declaration,Xml_comments,X_rdf],		%XML declaration + comments + RDF
	Xml_declaration=xmldecl([version='1.0']),		%XML version
	Xml_comments=comment('generated by daplex2rdfs written by K Hui'),	%comments
	X_rdf=env('rdf:RDF',Attribs,X_classes),			%RDF environment
	url(rdf,Rdf_url),
	url(rdfs,Rdfs_url),
	concat_atoms:concat_atoms([Rdf_url,'#'],Rdf_url_with_hash),
	concat_atoms:concat_atoms([Rdfs_url,'#'],Rdfs_url_with_hash),
	Attribs=[	'xmlns:rdf'=Rdf_url_with_hash,
			'xmlns:rdfs'=Rdfs_url_with_hash
		],
	findall(X,getentity(entmeta,X),All_classes),		%find all Daplex classes
	map_class(All_classes,X_classes),			%map all classes into RDFS
	print_xml:print_xml(Xml).

%
% map a list of Daplex classes into RDF schema definitions
%

%
% empty list
%
map_class([],[]) :-
	!.

%
% a list of classes
%
map_class([C|Cs],[R|Rs]) :-
	!,
	map_class(C,R),
	map_class(Cs,Rs).

%
% a class object, w/ superclass
%
map_class(Class,Rdf) :-
	getfnval(supertype,[Class],Superclass),					%get superclass
	getfnval(oname,[Superclass],Superclass_name),
	!,
	%
	% RDFS entry of the class
	%
	Rdf=[Rdf_comment,Rdf_class|Rdf_properties],
	%
	% the comment
	%
	Rdf_comment=comment(Class_comment),
	getfnval(oname,[Class],Class_name),
	concat_atoms:concat_atoms(['the entity class \"',Class_name,'\" in Daplex'],Class_comment),
	%
	% rdf:Description
	%
	Rdf_class=env('rdf:Description',Attribs,[Rdf_type,Rdfs_subclass_of]),
	Attribs=['rdf:ID'=Class_name],
	Rdf_type=elem('rdf:type',['rdf:resource'=Rdf_schema_class_url]),
	Rdfs_subclass_of=elem('rdfs:subClassOf',['rdf:resource'=Superclass_url]),
	url(rdfs,Rdf_schema_url),
	concat_atoms:concat_atoms([Rdf_schema_url,'#','Class'],Rdf_schema_class_url),
	concat_atoms:concat_atoms(['#',Superclass_name],Superclass_url),
	map_property(Class,Rdf_properties).

%
% a class object, w/ no superclass
%
map_class(Class,Rdf) :-
	%
	% RDFS entry of the class
	%
	Rdf=[Rdf_comment,Rdf_class,Rdf_properties],
	%
	% the comment
	%
	Rdf_comment=comment(Class_comment),
	getfnval(oname,[Class],Class_name),
	concat_atoms:concat_atoms(['the entity class ',Class_name,' in Daplex'],Class_comment),	%make up comment
	%
	% rdf:Description
	%
	Rdf_class=env('rdf:Description',Attribs,[Rdf_type,Rdfs_subclass_of]),
	Attribs=['rdf:ID'=Class_name],
	Rdf_type=elem('rdf:type',['rdf:resource'=Rdf_schema_class_url]),
	Rdfs_subclass_of=elem('rdfs:subClassOf',['rdf:resource'=Rdf_schema_resource_url]),
	url(rdfs,Rdf_schema_url),
	concat_atoms:concat_atoms([Rdf_schema_url,'#','Class'],Rdf_schema_class_url),
	concat_atoms:concat_atoms([Rdf_schema_url,'#','Resource'],Rdf_schema_resource_url),
	map_property(Class,Rdf_properties).

%
% map all properties of a class
%
map_property(Class,Rdf) :-
	findall(X,class_property(Class,X),Properties),
	map_property(Class,Properties,Rdf).

map_property(_,[],[]).
map_property(Class,[P|Ps],[Rdf_comment,Rdf_description|Rdfs]) :-
%	Rdf=[Rdf_comment,Rdf_description],
	getfnval(oname,[Class],Class_name),
	getfnval(fname,[P],Property_name),
	concat_atoms:concat_atoms(['property \"',Property_name,'\" on entity class \"',Class_name,'\"'],Comment),
	Rdf_comment=comment(Comment),
	Rdf_description=env('rdf:Description',['rdf:ID'=Id],[X_rdf_type,X_rdfs_domain,X_rdfs_range]),
	X_rdf_type=elem('rdf:type',['rdf:resource'=Type]),
	X_rdfs_domain=elem('rdfs:domain',['rdf:resource'=Domain]),
	X_rdfs_range=elem('rdfs:range',['rdf:resource'=Range]),
	getfnval(fname,[P],Property_name),
	getfnval(oname,[Class],Class_name),
	concat_atoms:concat_atoms([Class_name,'_',Property_name],Id),
	url(rdf,Rdf_syntax_url),
	concat_atoms:concat_atoms([Rdf_syntax_url,'#Property'],Type),	%type is rdf property
	concat_atoms:concat_atoms(['#',Class_name],Domain),			%domain is class name
	get_property_range(P,Range),
	map_property(Class,Ps,Rdfs).

get_property_range(P,Range) :-
	getfnval(result_type,[P],Result_objmeta),		%get result type of property
	relative(entmeta,Result_objmeta,Result_entmeta),	%cast it to entmeta
	getfnval(oname,[Result_entmeta],Type_name),		%get name of result type
	concat_atoms:concat_atoms(['#',Type_name],Range).

get_property_range(P,Range) :-
	getfnval(result_type,[P],Result_objmeta),		%get result type of property
	relative(simplemeta,Result_objmeta,_),			%cast it to simplemeta
	url(rdfs,Rdf_schema_url),
	concat_atoms:concat_atoms([Rdf_schema_url,'#Literal'],Range).

%
% find properties of a class, which is to appear in the RDFS
%
class_property(Class,Function) :-
	getentity(funmeta,Function),					%find a funmeta object
	not_inverse_function(Function),					%check that this is not a P/FDM derived inverse function
%	not_key(Class,Function),						%property is not a key
	getfnval(first_fun_arg,[Function],Argument),			%find its 1st argument
	relative(entmeta,Argument,Class).				%which is a entmeta

%
% make sure that function is not an inverse functin created by P/FDM
% we don't need this in RDF
%
not_inverse_function(Function) :-
	getfnval(fname,[Function],Fname),		%get function name
	name(Fname,Fname_string),			%convert to a string
	append(_,"_inv",Fname_string),			%has the suffix "_inv"
	!,
	fail.						%fail!
not_inverse_function(_).

not_key(Class,Function) :-
	getfnval(fname,[Function],Fname),		%get function name
	getfnval(oname,[Class],Class_name),		%get class name
	edescj(Class_name,_,_,Key_list,_,_,_),
	member(Fname,Key_list),
	!,
	fail.

not_key(_,_).


/*
%
% concat a list of nested atoms
%
concat_atoms([],X) :-
	name(X,[]).
concat_atoms([A|As],Atom) :-
	name(A,S),
	concat_atoms(As,Temp_a),
	name(Temp_a,Ss),
	append(S,Ss,String),
	name(Atom,String).
*/
:- write('***'),nl.
:- write('***'),nl.
:- write('*** REMEMBER: Please load in the database schema before calling go/0.'),nl.
:- write('***'),nl.
:- write('***'),nl.
