%
% compiling Daplex schema into RDFS
%
% By: K Hui
% Date: 12-4-2002
%

:- ['../../../prolog/pillow/pillow'].		%load pillow
:- ['../../../prolog/util/print_xml'].
:- ['../../../prolog/util/concat_atoms'].

%:- daplex('cif.daplex').			%load Daplex schema of CIF

%
% known URLs
%
:- dynamic(url/2).

url(rdfs,'http://www.w3.org/2000/01/rdf-schema').
url(rdf,'http://www.w3.org/1999/02/22-rdf-syntax-ns').
url(cifmet,'http://www.csd.abdn.ac.uk/~khui/akt/cif/cif-compact-rdfs.xml').
url(querymet,'http://www.csd.abdn.ac.uk/~khui/akt/cif/query.xml').

%
% map all modules
%
go :-
	Goal=(	getentity(modmeta,M),
		getfnval(mname,[M],Name)
		),
	findall(Name,call(Goal),All_module_names),		%find all modules
	map_modules(All_module_names).

%
% map 1 particular module
%
go(Module_name) :-
	map_module(Module_name).

%
% map a list of module, given their names
%
map_modules([]).
map_modules([N|Ns]) :-
	map_module(N),			%map 1 module
	map_modules(Ns).		%map remaining modules

%
% given the module name, map the module to RDFS
%
map_module(Module_name) :-
	Xml=[Xml_declaration,Xml_comments,X_rdf],		%XML declaration + comments + RDF
	Xml_declaration=xmldecl([version='1.0']),		%XML version
	Xml_comments=comment('generated by daplex2rdfs written by K Hui'),	%comments
	X_rdf=env('rdf:RDF',Attribs,X_classes),			%RDF environment
	url(rdf,Rdf_url),
	url(rdfs,Rdfs_url),
	concat_atoms:concat_atoms([Rdf_url,'#'],Rdf_url_with_hash),
	concat_atoms:concat_atoms([Rdfs_url,'#'],Rdfs_url_with_hash),
	Attribs=[	'xmlns:rdf'=Rdf_url_with_hash,
			'xmlns:rdfs'=Rdfs_url_with_hash
		],
	getentity(modmeta,[Module_name],Module),
	find_module_classes(Module,All_classes),
	map_classes(Module,All_classes,X_classes),		%map all classes into RDFS
	print_xml:print_xml(Xml).

%
% find all classes in a module
%
find_module_classes(Module,Classes) :-
	Goal=(	getentity(entmeta,Class),
		getfnval(cmodule,[Class],Module)
		),
	findall(Class,call(Goal),Classes).

%
% map a list of Daplex classes into RDF schema definitions
%

%
% empty list
%
map_classes(_,[],[]) :-
	!.

%
% a list of classes
%
map_classes(Module,[C|Cs],[R|Rs]) :-
	!,
	map_class(Module,C,R),
	map_classes(Module,Cs,Rs).

%
% a class object, w/ superclass
%
map_class(Module,Class,Rdf) :-
	getfnval(supertype,[Class],Superclass),					%get superclass
	!,
	%
	% RDFS entry of the class
	%
	Rdf=[Rdf_comment,Rdf_class|Rdf_properties],
	%
	% the comment
	%
	Rdf_comment=comment(Class_comment),
	getfnval(oname,[Class],Class_name),
	concat_atoms:concat_atoms(['the entity class \"',Class_name,'\" in Daplex'],Class_comment),
	%
	% rdf:Description
	%
	Rdf_class=env('rdf:Description',Attribs,[Rdf_type,Rdfs_subclass_of]),
	Attribs=['rdf:ID'=Class_name],
	Rdf_type=elem('rdf:type',['rdf:resource'=Rdf_schema_class_url]),
	Rdfs_subclass_of=elem('rdfs:subClassOf',['rdf:resource'=Superclass_url]),
	url(rdfs,Rdf_schema_url),
	concat_atoms:concat_atoms([Rdf_schema_url,'#','Class'],Rdf_schema_class_url),
	compose_class_url(Module,Superclass,Superclass_url),
	map_property(Module,Class,Rdf_properties).

%
% a class object, w/ no superclass
%
map_class(Module,Class,Rdf) :-
	%
	% RDFS entry of the class
	%
	Rdf=[Rdf_comment,Rdf_class,Rdf_properties],
	%
	% the comment
	%
	Rdf_comment=comment(Class_comment),
	getfnval(oname,[Class],Class_name),
	concat_atoms:concat_atoms(['the entity class ',Class_name,' in Daplex'],Class_comment),	%make up comment
	%
	% rdf:Description
	%
	Rdf_class=env('rdf:Description',Attribs,[Rdf_type,Rdfs_subclass_of]),
	Attribs=['rdf:ID'=Class_name],
	Rdf_type=elem('rdf:type',['rdf:resource'=Rdf_schema_class_url]),
	Rdfs_subclass_of=elem('rdfs:subClassOf',['rdf:resource'=Rdf_schema_resource_url]),
	url(rdfs,Rdf_schema_url),
	concat_atoms:concat_atoms([Rdf_schema_url,'#','Class'],Rdf_schema_class_url),
	concat_atoms:concat_atoms([Rdf_schema_url,'#','Resource'],Rdf_schema_resource_url),
	map_property(Module,Class,Rdf_properties).

%
% compose URL of a class
%
compose_class_url(_,Class,Class_url) :-
	getentity(simplemeta,Class),						%is a simplemeta
	url(rdfs,Rdf_schema_url),						%get RDFS URL
	concat_atoms:concat_atoms([Rdf_schema_url,'#Literal'],Class_url),	%just a RDFS literal
	!.

compose_class_url(_,Class,Class_url) :-
	relative(simplemeta,Class,_),						%can be casted into simplemeta
	url(rdfs,Rdf_schema_url),						%get RDFS URL
	concat_atoms:concat_atoms([Rdf_schema_url,'#Literal'],Class_url),	%just a RDFS literal
	!.

compose_class_url(Module,Class,Class_url) :-
	relative(compoundmeta,Class,Compound_class),			%cast it into compoundmeta in order to get module info.
	getfnval(cmodule,[Compound_class],Module),			%class is in current module
	getfnval(oname,[Class],Class_name),				%get class name
	concat_atoms:concat_atoms(['#',Class_name],Class_url),
	!.

compose_class_url(_,Class,Class_url) :-
	relative(compoundmeta,Class,Compound_class),			%cast into compoundmeta
	getfnval(cmodule,[Compound_class],Class_module),		%get class module
	getfnval(mname,[Class_module],Class_module_name),		%get name of class module
	getfnval(oname,[Class],Class_name),
	url(Class_module_name,Class_module_rdfs_url),
	concat_atoms([Class_module_rdfs_url,'#',Class_name],Class_url),
	!.

compose_class_url(_,Class,_) :-
	getfnval(oname,[Class],Class_name),
	write('*** daplex2rdfs ERROR: cannot compose URL of class:'),
	write(Class_name),
	nl,
	fail.

%
% map all properties of a class
%
map_property(Module,Class,Rdf) :-
	findall(X,class_property(Class,X),Properties),
	map_property(Module,Class,Properties,Rdf).

map_property(_,_,[],[]).
map_property(Module,Class,[P|Ps],[Rdf_comment,Rdf_description|Rdfs]) :-
	getfnval(oname,[Class],Class_name),				%get class name
	getfnval(fname,[P],Property_name),				%get property name
	concat_atoms:concat_atoms(['property \"',Property_name,'\" on entity class \"',Class_name,'\"'],Comment),
	Rdf_comment=comment(Comment),
	Rdf_description=env('rdf:Description',['rdf:ID'=Id],[X_rdf_type,X_rdfs_domain,X_rdfs_range]),
	X_rdf_type=elem('rdf:type',['rdf:resource'=Type]),
	X_rdfs_domain=elem('rdfs:domain',['rdf:resource'=Domain_url]),
	X_rdfs_range=elem('rdfs:range',['rdf:resource'=Range_url]),
	concat_atoms:concat_atoms([Class_name,'_',Property_name],Id),	%compose property name
	url(rdf,Rdf_syntax_url),					%get RDF URL
	concat_atoms:concat_atoms([Rdf_syntax_url,'#Property'],Type),	%type is rdf property
	%
	% compose domain & range URL
	%
	compose_class_url(Module,Class,Domain_url),			%get domain URL
%	concat_atoms:concat_atoms(['#',Class_name],Domain),		%domain is class name
%	get_property_range(Module,P,Range),
	getfnval(result_type,[P],Range_type),				%get result type of property
	compose_class_url(Module,Range_type,Range_url),			%get domain URL
	map_property(Module,Class,Ps,Rdfs).

/*
%
% property is a compound type
%
get_property_range(Module,P,Range) :-
	getfnval(result_type,[P],Result_type),				%get result type of property
	compose_class_url(Module,Result_type,Range).			%compose URL

%
% property is a simple type
%
get_property_range(_,P,Range) :-
	getfnval(result_type,[P],Result_objmeta),			%get result type of property
	relative(simplemeta,Result_objmeta,_),				%can be casted into simplemeta
	url(rdfs,Rdf_schema_url),					%get RDFS URL
	concat_atoms:concat_atoms([Rdf_schema_url,'#Literal'],Range).	%just a RDFS literal
*/

%
% find properties of a class, which is to appear in the RDFS
%
class_property(Class,Function) :-
	getentity(funmeta,Function),					%find a funmeta object
	not_inverse_function(Function),					%check that this is not a P/FDM derived inverse function
%	not_key(Class,Function),						%property is not a key
	getfnval(first_fun_arg,[Function],Argument),			%find its 1st argument
	relative(entmeta,Argument,Class).				%which is a entmeta

%
% make sure that function is not an inverse functin created by P/FDM
% we don't need this in RDF
%
not_inverse_function(Function) :-
	getfnval(fname,[Function],Fname),		%get function name
	name(Fname,Fname_string),			%convert to a string
	append(_,"_inv",Fname_string),			%has the suffix "_inv"
	!,
	fail.						%fail!
not_inverse_function(_).

not_key(Class,Function) :-
	getfnval(fname,[Function],Fname),		%get function name
	getfnval(oname,[Class],Class_name),		%get class name
	edescj(Class_name,_,_,Key_list,_,_,_),
	member(Fname,Key_list),
	!,
	fail.

not_key(_,_).


:- write('***'),nl.
:- write('***'),nl.
:- write('*** REMEMBER: Please load in the database schema before calling go/0.'),nl.
:- write('***'),nl.
:- write('***'),nl.


%:- daplex('cif-compact.daplex').
%:- daplex('cif-classic.daplex').
%:- daplex('cif-query.daplex').
