%
% In this file, we specify how different classes are represented as Prolog terms.
% We also tell the system how to get a Property's Prolog term from a Instance, and vice versa.
%
% class & property Prolog term in compact CIF
%
% a class w/ its property
%
:- ['../util/nth'].

%cif_node cif_node_id

%class_property(cif_compact,objmet,objmet_oname,_,Cterm,Pterm) :-

%
% entmet
%
%entmet entmet_super
%entmet entmet_rdfname
%

%---entmet/entmet_rdfname

%
% Cterm->Pterm
%
class_property(cif_compact,entmet,entmet_rdfname,Params,Cterm,Pterm) :-
	var(Pterm),
	!,
	atom(Cterm),
	member(domain_rdfs=Domain_url,Params),				%get domain URL from params
	concat_atoms:concat_atoms([Domain_url,'#',Cterm],Pterm).

%
% Pterm->Cterm
%
class_property(cif_compact,entmet,entmet_rdfname,Params,Cterm,Pterm) :-
	member(domain_rdfs=Domain_url,Params),				%get domain URL from params
	concat_atoms:add_prefix(Temp,Cterm,Pterm),
	concat_atoms:add_prefix(Domain_url,'#',Temp).

%
% propmet
%
%propmet propmet_fname
%propmet propmet_firstargtype
%propmet propmet_resulttype
%propmet propmet_rdfname
%


%---propmet/propmet_fname

%
% Cterm<->Pterm
%
class_property(cif_compact,propmet,propmet_fname,_,Cterm,Pterm) :-
	Cterm=generate(Pterm,_,_,_,_).


%---propmet/propmet_firstargtype

%
% Cterm->Pterm
%
class_property(cif_compact,propmet,propmet_firstargtype,_,Cterm,Pterm) :-
	var(Pterm),
	!,
	Cterm=generate(_,[Pterm|_],_,_,_).

%
% Pterm->Cterm
%
class_property(cif_compact,propmet,propmet_firstargtype,_,Cterm,Pterm) :-
	Cterm=generate(_,[Pterm],_,_,_).


%---propmet/propmet_resulttype

%
% Cterm->Pterm
%
class_property(cif_compact,propmet,propmet_resulttype,_,Cterm,Pterm) :-
	Cterm=generate(_,_,_,Pterm,_).


%---propmet/propmet_rdfname

%
% Cterm->Pterm
%
class_property(cif_compact,propmet,propmet_rdfname,Params,Cterm,Pterm) :-
	var(Pterm),
	!,
	Cterm=generate(Fname,_,_,_,_),
	member(domain_rdfs=Domain_url,Params),
	concat_atoms:concat_atoms([Domain_url,'#',Fname],Pterm).

%
% Pterm->Cterm
%
class_property(cif_compact,propmet,propmet_rdfname,Params,Cterm,Pterm) :-
	member(domain_rdfs=Domain_url,Params),
	concat_atoms:add_prefix(Temp,Fname,Pterm),
	concat_atoms:add_prefix(Domain_url,'#',Temp),
	Cterm=generate(Fname,_,_,_,_).


%class_property(cif_compact,constraint,constraint_id,_,Cterm,Pterm) :-

%
% variable
%
%variable variable_varname
%

%---variable/variable_varname

%
% Cterm<->Pterm
%
class_property(cif_compact,variable,variable_varname,_,Cterm,Pterm) :-
	Cterm=var(Pterm).

%
% arith_expr
%
%arith_expr arith_expr_operator
%arith_expr arith_expr_operand1
%arith_expr arith_expr_operand2
%arith_expr arith_expr_result
%

%---arith_expr/arith_expr_operator

%
% Cterm<->Pterm
%
class_property(cif_compact,arith_expr,arith_expr_operator,_,Cterm,Pterm) :-
	Cterm=expression(_,_,expr(Pterm,_,_)).

%---arith_expr/arith_expr_operand1

%
% Cterm->Pterm
%
class_property(cif_compact,arith_expr,arith_expr_operand1,_,Cterm,Pterm) :-
	var(Pterm),
	!,
	Cterm=expression(_,_,expr(_,Pterm,_)).

%
% Pterm->Cterm
%
class_property(cif_compact,arith_expr,arith_expr_operand1,_,Cterm,Pterm) :-
	Cterm=expression(_,Req_vars,expr(_,Pterm,_)),
	var(Req_vars),						%Req_vars is not instantiated at all
	!,
	(	(	Pterm=var(_),				%Pterm is a variable
			Req_vars=[Pterm|_]			%instantiated the 1st element, leave the tail
		);
		true						%Pterm is NOT a variable, just leave it
	),
	!.

class_property(cif_compact,arith_expr,arith_expr_operand1,_,Cterm,Pterm) :-
	Cterm=expression(_,Req_vars,expr(_,Pterm,_)),
	Req_vars=[_|Tail],					%head is instantiated
	var(Tail),						%tail is not
	(	(	Pterm=var(_),				%now Pterm is a Daplex var
			Tail=[Pterm]				%put it into the tail
		);
		Tail=[]						%otherwise forget it
	),
	!.


%---arith_expr/arith_expr_operand2

%
% Cterm->Pterm
%
class_property(cif_compact,arith_expr,arith_expr_operand2,_,Cterm,Pterm) :-
	Cterm=expression(_,_,expr(_,_,Pterm)).

%
% Pterm->Cterm
%
class_property(cif_compact,arith_expr,arith_expr_operand2,_,Cterm,Pterm) :-
	Cterm=expression(_,Req_vars,expr(_,_,Pterm)),
	var(Req_vars),						%Req_vars is not instantiated at all
	!,
	(	(	Pterm=var(_),				%Pterm is a variable
			Req_vars=[Pterm|_]			%instantiated the 1st element, leave the tail
		);
		true						%Pterm is NOT a variable, just leave it
	),
	!.

class_property(cif_compact,arith_expr,arith_expr_operand2,_,Cterm,Pterm) :-
	Cterm=expression(_,Req_vars,expr(_,_,Pterm)),
	Req_vars=[_|Tail],					%head is instantiated
	var(Tail),						%tail is not
	(	(	Pterm=var(_),				%now Pterm is a Daplex var
			Tail=[Pterm]				%put it into the tail
		);
		Tail=[]						%otherwise forget it
	),
	!.


%---arith_expr/arith_expr_result

%
% Cterm->Pterm
%
class_property(cif_compact,arith_expr,arith_expr_result,_,Cterm,Pterm) :-
	var(Pterm),
	Cterm=expression([Pterm],_,expr(_,_,_)),
	!.

class_property(cif_compact,arith_expr,arith_expr_result,_,Cterm,Pterm) :-
	var(Pterm),
	!,
	Cterm=expression(Pterm,_,expr(_,_,_)).

%
% Pterm->Cterm
%
class_property(cif_compact,arith_expr,arith_expr_result,_,Cterm,Pterm) :-
	Cterm=expression([Pterm],_,expr(_,_,_)),
	!.




%
% casting_expr
%
%casting_expr casting_expr_casting_class
%casting_expr casting_expr_casting_from
%casting_expr casting_expr_casting_to
%

%---casting_expr/casting_expr_casting_class

%
% Cterm<->Pterm
%
class_property(cif_compact,casting_expr,casting_expr_casting_class,_,Cterm,Pterm) :-
	Cterm=relative(Pterm,_,_).

%---casting_expr/casting_expr_casting_from

%
% Cterm<->Pterm
%
class_property(cif_compact,casting_expr,casting_expr_casting_from,_,Cterm,Pterm) :-
	Cterm=relative(_,Pterm,_).

%---casting_expr/casting_expr_casting_to

%
% Cterm<->Pterm
%
class_property(cif_compact,casting_expr,casting_expr_casting_to,_,Cterm,Pterm) :-
	Cterm=relative(_,_,Pterm).


%
% booland
%
% booland/booland_op
%

%---booland/booland_op

%
% Cterm->Pterm
%
class_property(cif_compact,booland,booland_op,_,Cterm,Pterm) :-
	var(Pterm),
	Cterm=logical_and(Pterm,_).			%Cterm is a logical_and/2

class_property(cif_compact,booland,booland_op,_,Cterm,Pterm) :-
	var(Pterm),
	Cterm=logical_and(_,Pterm).			%Cterm is a logical_and/2

class_property(cif_compact,booland,booland_op,_,Cterm,Pterm) :-
	var(Pterm),
	member(Pterm,Cterm).				%Cterm is a list

%
% *** this prevent calling "Pterm->Cterm" clauses when we backtrack
% *** we need to allow backtracking in clauses above to get alternate
% *** values of booland_op in a booland (which is a list)
%
class_property(cif_compact,booland,booland_op,_,_,Pterm) :-
	var(Pterm),
	!,
	fail.

%
% Pterm->Cterm
%
class_property(cif_compact,booland,booland_op,_,Cterm,Pterm) :-
%	var(Cterm),
	find_first_var(Cterm,First_var),		%find 1st variable element in the list & unify it w/ Pterm
	First_var=Pterm,
	!.


%---boolor/boolor_op
%
% Cterm<->Pterm
%
class_property(cif_compact,boolor,boolor_op1,_,Cterm,Pterm) :-
	Cterm=logical_or(Pterm,_).

class_property(cif_compact,boolor,boolor_op2,_,Cterm,Pterm) :-
	Cterm=logical_or(_,Pterm).


%---negpred/negpred_negop

%
% Cterm<->Pterm
%
class_property(cif_compact,negpred,negpred_negop,_,Cterm,'-') :-
	Cterm=icode_not(_).

%---negpred/negpred_body

%
% Cterm<->Pterm
%
class_property(cif_compact,negpred,negpred_body,_,Cterm,Pterm) :-
	Cterm=icode_not(Pterm).


%---setmem/setmem_var

%
% Cterm<->Pterm
%
class_property(cif_compact,setmem,setmem_var,_,Cterm,Pterm) :-
	Cterm=gen(Pterm,_,_).

%---setmem/setmem_set

%
% Cterm->Pterm
%
class_property(cif_compact,setmem,setmem_set,_,Cterm,Pterm) :-
	var(Pterm),
	Cterm=gen(_,_,[Pterm|_]),
	!.

class_property(cif_compact,setmem,setmem_set,_,Cterm,Pterm) :-
	var(Pterm),
	Cterm=gen(_,[Pterm|_],[]),
	!.

%
% Pterm->Cterm
%
class_property(cif_compact,setmem,setmem_set,_,Cterm,Pterm) :-
	Pterm=generate(_,Var),
	!,
	Cterm=gen(Var,[Pterm],Gen5),
	(	(	var(Gen5),
			Gen5=[]
		);
		true
	),
	!.

class_property(cif_compact,setmem,setmem_set,_,Cterm,Pterm) :-
	Pterm=generate(_,_,_,_,Var),
	!,
	Cterm=gen(Var,Gen2,[Pterm]),
	(	(	var(Gen2),
			Gen2=[]
		);
		true
	),
	!.


%---mvfncall/mvfncall_prop

%
% mvfncall
%
%mvfncall mvfncall_prop
%mvfncall mvfncall_arg
%
class_property(cif_compact,mvfncall,mvfncall_prop,_,Cterm,Pterm) :-
	Cterm=generate(_,_,_,_,_),
	Cterm=Pterm.

class_property(cif_compact,mvfncall,mvfncall_arg,_,Cterm,Pterm) :-
	Cterm=generate(_,_,_,_,_),
	Cterm=Pterm.

%
% entset
%
%entset entset_entclass
%
class_property(cif_compact,entset,entset_entclass,_,Cterm,Pterm) :-
	Cterm=generate(Pterm,_).

%
% int_set
%
%int_set int_set_start
%int_set int_set_end
%int_set int_set_var
%
class_property(cif_compact,int_set,int_set_start,_,Cterm,Pterm) :-
	Cterm=expression(_,_,int_set(Pterm,_,_)).

class_property(cif_compact,int_set,int_set_end,_,Cterm,Pterm) :-
	Cterm=expression(_,_,int_set(_,Pterm,_)).

class_property(cif_compact,int_set,int_set_var,_,Cterm,Pterm) :-
	Cterm=expression(Pterm,_,int_set(_,_,_)).

%
% explicit_set
%
%explicit_set explicit_set_element
%
class_property(cif_compact,explicit_set,explicit_set_element,_,Cterm,Pterm) :-
	Cterm=generate_subquery(_,_,explicit(Pterm),_,_).





%
% comparison
%
%comparison comparison_operator
%comparison comparison_op1
%comparison comparison_op2
%

%---comparison/comparison_operator


class_property(cif_compact,comparison,comparison_operator,_,Cterm,Pterm) :-
	Cterm=expression(_,_,expr(Op,_,_)),
	comparison_operator(Op,Pterm).

%---comparison/comparison_op1

%
% Cterm->Pterm
%
class_property(cif_compact,comparison,comparison_op1,_,Cterm,Pterm) :-
	var(Pterm),
	!,
	Cterm=expression(_,_,expr(_,Pterm,_)).

%
% Pterm->Cterm
%
% case 1: Req_vars list is not instantiated yet
%
class_property(cif_compact,comparison,comparison_op1,_,Cterm,Pterm) :-
	Cterm=expression(Result_var,Req_vars,expr(_,Pterm,_)),
	%
	% try instantiating Result_var & Req_vars
	% Req_vars is uninstantiated, put Pterm into 1st element
	%
	var(Req_vars),						%Req_vars is not instantiated at all
	!,
	(	(	Pterm=var(_),				%Pterm is a variable
			Req_vars=[Pterm|_]			%instantiated the 1st element, leave the tail
		);
		true						%Pterm is NOT a variable, just leave it
	),
	(	(	var(Result_var),
			Result_var=[]
		);
		true
	),
	!.

%
% case 2: Req_vars list is partially instantiated
%
class_property(cif_compact,comparison,comparison_op1,_,Cterm,Pterm) :-
	Cterm=expression(Result_var,Req_vars,expr(_,Pterm,_)),
	%
	% try instantiating Result_var & Req_vars
	%
	Req_vars=[First|Rest],				%get 1st & tail
	nonvar(First),					%1st element is instantiated
	var(Rest),					%tail is not
	!,
	(	(	Pterm=var(_),			%Pterm is a variable
			Rest=[Pterm]			%instantiated the tail by Pterm
		);
		Rest=[]
	),
	(	(	var(Result_var),
			Result_var=[]
		);
		true
	),
	!.


%---comparison/comparison_op2

%
% Cterm->Pterm
%
class_property(cif_compact,comparison,comparison_op2,_,Cterm,Pterm) :-
	var(Pterm),
	!,
	Cterm=expression(_,_,expr(_,_,Pterm)).

%
% Pterm->Cterm
%
class_property(cif_compact,comparison,comparison_op2,_,Cterm,Pterm) :-
	Cterm=expression(Result_var,Req_vars,expr(_,_,Pterm)),
	%
	% try instantiating Result_var & Req_vars
	% Req_vars is uninstantiated, put Pterm into 1st element
	%
	var(Req_vars),						%Req_vars is not instantiated at all
	!,
	(	(	Pterm=var(_),				%Pterm is a variable
			Req_vars=[Pterm|_]			%instantiated the 1st element, leave the tail
		);
		true						%Pterm is NOT a variable, just leave it
	),
	(	(	var(Result_var),
			Result_var=[]
		);
		true
	),
	!.

class_property(cif_compact,comparison,comparison_op2,_,Cterm,Pterm) :-
	Cterm=expression(Result_var,Req_vars,expr(_,_,Pterm)),
	%
	% try instantiating Result_var & Req_vars
	%
	Req_vars=[First|Rest],				%get 1st & tail
	nonvar(First),					%1st element is instantiated
	var(Rest),					%tail is not
	!,
	(	(	Pterm=var(_),			%Pterm is a variable
			Rest=[Pterm]			%instantiated the tail by Pterm
		);
		Rest=[]
	),
	(	(	var(Result_var),
			Result_var=[]
		);
		true
	),
	!.

%
% impliesconstr
%
%impliesconstr impliesconstr_qvar
%impliesconstr impliesconstr_if
%impliesconstr impliesconstr_then
%

%---impliesconstr/impliesconstr_qvar

%
% Cterm->Pterm
%
class_property(cif_compact,impliesconstr,impliesconstr_qvar,_,Cterm,Pterm) :-
	var(Pterm),
	!,
	Cterm=restrict_subquery(all(Var),Gen,_),
	Patterns=[generate(_,Var),generate(_,_,_,_,Var)],	%pattern that matches generate/2 & generate/5
	match_pattern(Gen,Patterns,Generators,_),		%get generators of this var & extra condition
	%
	% divide it into generate/2 & generate/5
	%
	Gen2_pattern=generate(_,Var),
	match_pattern(Generators,Gen2_pattern,Gen2,Gen5),		%get generators of this var & extra condition
	Pterm=gen(Var,Gen2,Gen5).

%
% Pterm->Cterm
%
class_property(cif_compact,impliesconstr,impliesconstr_qvar,_,Cterm,Pterm) :-
	Pterm=gen(Var,Gen2,Gen5),			%get gen2 & gen5 from qvar
	append(Gen2,Gen5,Temp),
	unify_var_in_gen(Var,Temp),			%unify variable w/ all gen/2 & gen/5
	append(Temp,_If,Gen),				%leave IF part uninstantiated
	Cterm=restrict_subquery(all(Var),Gen,_).


%
% Cterm->Pterm
%
class_property(cif_compact,impliesconstr,impliesconstr_if,_,Cterm,Pterm) :-
	var(Pterm),
	!,
	Cterm=restrict_subquery(all(Var),Gen,_),
	Patterns=[generate(_,Var),generate(_,_,_,_,Var)],	%pattern that matches generate/2 & generate/5
	match_pattern(Gen,Patterns,_,Condition),		%get generators of this var & extra condition
	(	(	Condition=[],				%if condition is empty, take it as a true
			!,
			If_to_parse=icode_true
		);
		If_to_parse=Condition
	),
	Pterm=If_to_parse.

%
% Pterm->Cterm
%
class_property(cif_compact,impliesconstr,impliesconstr_if,_,Cterm,Pterm) :-
	Cterm=restrict_subquery(all(_),Gen,_),		%get gen
	append(Gen2_and_5,If,Gen),			%get the IF part
	var(If),					%which is uninstantiated
	(	(	Pterm=icode_true,
			If=[]
		);
		(	Pterm=[icode_true],
			If=[]
		);
		(	Pterm=[_|_],
			If=Pterm
		);
		If=[Pterm]
	),
	!.



%---impliesconstr/impliesconstr_then

%
% Cterm->Pterm
%
class_property(cif_compact,impliesconstr,impliesconstr_then,_,Cterm,Pterm) :-
	var(Pterm),
	!,
	Cterm=restrict_subquery(all(_),_,Pred),
	(	(	Pred=[],				%if pred is empty, take it as a true
			!,
			Then_to_parse=icode_true
		);
		Then_to_parse=Pred
	),
	Pterm=Then_to_parse.

%
% Pterm->Cterm
%
class_property(cif_compact,impliesconstr,impliesconstr_then,_,Cterm,Pterm) :-
	Cterm=restrict_subquery(all(_),_,Pterm).


%
% existsconstr
%
%existsconstr existsconstr_qvar
%existsconstr existsconstr_suchthat
%
class_property(cif_compact,existsconstr,existsconstr_qvar,_,Cterm,Pterm) :-
	Cterm=restrict_subquery(some(Var),Gen,_),
	Patterns=[generate(_,Var),generate(_,_,_,_,Var)],	%pattern for matching generate/2 & /5
	match_pattern(Gen,Patterns,Generators,_),		%get generators & non-gen
	Pterm=gen(Var,Generators).

class_property(cif_compact,existsconstr,existsconstr_suchthat,_,Cterm,Pterm) :-
	Cterm=restrict_subquery(some(Var),Gen,Pred),
	Patterns=[generate(_,Var),generate(_,_,_,_,Var)],	%pattern for matching generate/2 & /5
	match_pattern(Gen,Patterns,_,Non_gens),			%get generators & non-gen
	append_cif(Non_gens,Pred,Suchthat),			%append pred to non-gen
	(	(	Suchthat=[],				%if pred is empty, take it as a true
			!,
			Suchthat_to_parse=icode_true
		);
		Suchthat_to_parse=Suchthat
	),
	Pterm=Suchthat_to_parse.

%
% unquantified_constraint
%
%unquantified_constraint unquantified_constraint_body
%
class_property(cif_compact,unquantified_constraint,unquantified_constraint_body,_,Cterm,Cterm).

%
% recpred
%
%recpred recpred_body
%
class_property(cif_compact,recpred,recpred_body,_,Cterm,Cterm).

%
% simplemet
%
%simplemet simplemet_data_type
%
class_property(cif_compact,simplemet,simplemet_data_type,_,integer,integer).

class_property(cif_compact,simplemet,simplemet_data_type,_,float,float).

class_property(cif_compact,simplemet,simplemet_data_type,_,boolean,boolean).

class_property(cif_compact,simplemet,simplemet_data_type,_,string,string).

%
% constant
%

%
%boolconst boolconst_value
%
class_property(cif_compact,boolconst,boolconst_value,_,icode_true,'True') :-
	!.

class_property(cif_compact,boolconst,boolconst_value,_,icode_false,'False') :-
	!.

%
%stringconst stringconst_value
%
class_property(cif_compact,stringconst,stringconst_value,_,Cif,Cif) :-
	atom(Cif),
	!.

%
%integerconst integerconst_value
%
class_property(cif_compact,integerconst,integerconst_value,_,Cif,Cif) :-
	integer(Cif),
	!.

%
%floatconst floatconst_value
%
class_property(cif_compact,floatconst,floatconst_value,_,Cif,Cif) :-
	number(Cif),
	!.

%
% boolprim
%
%boolprim boolprim_value
%
class_property(cif_compact,boolprim,boolprim_value,_,Cif,Cif).




comparison_operator(Op,Xml) :-
	Cif_ops=[	'=',
			'<>',
			'<',
			'=<',
			'>',
			'>='
		],
	Xml_ops=[	'=',
			'&lt;&gt;',
			'&lt;',
			'=&lt;',
			'&gt;',
			'&gt;='
			],
	nth(Index,Cif_ops,Op),
	nth(Index,Xml_ops,Xml).


unify_var_in_gen(_,[]).
unify_var_in_gen(Var,[generate(_,Var)|Rest]) :-
	unify_var_in_gen(Var,Rest).

unify_var_in_gen(Var,[generate(_,_,Var,_,_)|Rest]) :-
	unify_var_in_gen(Var,Rest).

%
% find the 1st var element in a list
%
find_first_var([X|_],X) :-
	var(X),
	!.

find_first_var([_|X],First) :-
	find_first_var(X,First).

%
% rounding up predicate when mapping from triples to Prolog term
% everything which is uninstantiated got instantiated to the empty list
%
triple2prolog_roundup(cif_compact,[],[]) :-
	!.

triple2prolog_roundup(cif_compact,[B|Bs],[A|As]) :-
	triple2prolog_roundup(cif_compact,B,A),
	triple2prolog_roundup(cif_compact,Bs,As),
	!.
	
triple2prolog_roundup(cif_compact,Before,Before) :-
	functor(Before,F,0),
	!.

triple2prolog_roundup(cif_compact,Before,After) :-
	Before =.. Before_list,
	triple2prolog_roundup(cif_compact,Before_list,After_list),
	After =.. After_list.

