%
% compile a RDF file to triples and write it to a file
% this allow a fast loading of RDF triples into rdf2triple
% and avoid parsing the RDF every time
%

:- ['../pillow/pillow'].
:- ['../util/print_xml'].
:- [rdf2triple].
:- use_module(library(lists)).

go(Url,File) :-
	rdf2triple:reset,
	get_tree(Url,Xml_tree),
	get_rdf_tree(Xml_tree,Rdf_tree),
	rdf2triple:rdf2triple(Url,Rdf_tree),
	tell(File),
	listing(triple:rdf_content),
	told.

get_tree(Url,Xml_tree) :-
	pillow:url_info(Url,Url_as_term),				%convert URL into a term
        pillow:fetch_url(Url_as_term,[],Response),			%fetch URL
        member(status(success,_,_),Response),				%check for success
        member(content(Content_string),Response),
        pillow:xml2terms(Content_string,Xml_tree).			%parse into XML tree

get_rdf_tree(Xml_tree,Rdf_tree) :-
	Rdf_tree=elem('rdf:RDF',_),
	member(Rdf_tree,Xml_tree).

get_rdf_tree(Xml_tree,Rdf_tree) :-
	Rdf_tree=env('rdf:RDF',_,_),
	member(Rdf_tree,Xml_tree).

:- write('***'),nl.
:- write('*** type go(URL,Filename) to compile an RDF file into triples'),nl.
:- write('***'),nl.
