:- module(parse_url,[parse_url/2]).
:- [concat_atoms].

%
% atom -> term
%
parse_url(Atom,url(Host,Port,Doc)) :-
	nonvar(Atom),
	atom(Atom),
	name(Atom,String),
	http_url(Host,Port,Doc,String,[]).

%
% string -> term
%
parse_url(String,url(Host,Port,Doc)) :-
	nonvar(String),
	http_url(Host,Port,Doc,String,[]).

%
% term -> atom
%
parse_url(Atom,url(Host,80,doc(Doc,Anchor))) :-
	var(Atom),
	!,
	concat_atoms:concat_atoms(['http://',Host,'/',Doc,'#',Anchor],Atom).

parse_url(Atom,url(Host,80,Doc)) :-
	var(Atom),
	!,
	concat_atoms:concat_atoms(['http://',Host,'/',Doc],Atom).

parse_url(Atom,url(Host,Port,doc(Doc,Anchor))) :-
	var(Atom),
	!,
	concat_atoms:concat_atoms(['http://',Host,':',Port,'/',Doc,'#',Anchor],Atom).

parse_url(Atom,url(Host,Port,Doc)) :-
	var(Atom),
	concat_atoms:concat_atoms(['http://',Host,':',Port,'/',Doc],Atom).


http_url(Host,Port,Doc) -->
	"http://",
	internet_host(Host),
	optional_port(Port),
	http_document(Doc).

%
% parse an internet host
%
internet_host(Host) -->
	internet_host_char(C),
	internet_host_chars(Cs),
	{
	name(Host,[C|Cs])
	}.

%
% zero or more internet char
%
internet_host_chars([C|Cs]) -->
	internet_host_char(C),
	internet_host_chars(Cs).

internet_host_chars([]) -->
	"".

%
% char allowed to appear in an internet host name
%
internet_host_char(C) -->
	digit(C).		%digit
internet_host_char(C) -->
	loalpha(C).		%lower case
internet_host_char(C) -->
	upalpha(C).		%upper case
internet_host_char(C) -->
	misc_host_char(C).

%internet_host_char(0'-) -->
%	"-".			%underscore
%internet_host_char(0'.) -->
%	".".

%
% an optional port no. in an internet host
%
optional_port(Port) -->
	":", !,
	parse_integer(Port).

optional_port(80) -->		%the default port no. 80
	"".

%
% the document part of a URL
%
%http_document([0'/|Doc]) -->
%        "/",
%	!,
%        rest(Doc).


http_document(doc(Doc,Anchor)) -->
	"/",
	document(Doc),
	"#",
	anchor(Anchor),
	!.
%	no_hash_chars(Anchor_string),
%	{	name(Doc,Doc_string),
%		name(Anchor,Anchor_string)
%	}.

http_document(Doc) -->
	"/",
	document(Doc),
	!.

http_document('/') -->
	"".

%
% rest of the list
%
rest(S,S,[]).

loupalpha(C) -->
	loalpha(C),
	!.
loupalpha(C) -->
	upalpha(CU),
	{	C is CU+0'a-0'A
	}.

loalpha(C) -->
	[C],
	{	C >= 0'a,
		C =< 0'z
	}.

upalpha(C) -->
	[C],
	{	C >= 0'A,
		C =< 0'Z
	}.

digit(C) -->
	[C],
	{	C >= 0'0,
		C =< 0'9
	}.

misc_host_char(0'.) -->
	[0'.].		%dot
misc_host_char(0'_) -->
	[0'_].		%underscore
misc_host_char(0'-) -->
	[0'-].		%dash

misc_host_char(0'~) -->
	[0'~].		%tilde

%
% valid document name chars
%
doc_chars([C|Cs]) -->
	doc_char(C),
	doc_chars(Cs).

doc_chars([]) -->
	"".

%
% valid char in document name
%
doc_char(0'/) -->
	"/".

doc_char(C) -->
	loalpha(C).

doc_char(C) -->
	upalpha(C).

doc_char(C) -->
	digit(C).

doc_char(C) -->
	misc_host_char(C).

%
% document in a URL
%
document(Doc) -->
	doc_chars(String),
	{	name(Doc,String)
	}.

%
% the bit after the "#"
%
anchor(Anchor) -->
	doc_chars(String),
	{	name(Anchor,String)
	}.

