%
% map XCIF back into CIF
%
%

:- module(xcif2cif,[	xcif2cif/2		%transform XCIF -> CIF
		]).

:- ['../../prolog/util/cif_util'].
:- ['../../prolog/util/sort_precedence'].
:- ['../../prolog/util/flatten'].
:- use_module(library(lists)).

%
% map a XCIF constraint back into CIF
%

%xcif2cif(Xcif,Cif) :-
%	cif_util:find_var_dependency(Xcif,Dependency),				%find var dependency in a list of depend/2
%	sort_precedence:clean_precedence,
%	add_precedence_info(Dependency),
%	sort_precedence:sort(Btrees),						%sort vars according to dependency
%	sort_precedence:btree2list(Btrees,Vars_list),				%map btrees into list of list
%	flatten:flatten(Vars_list,Vars),					%flatten list of list into a list
%	xcif2cif([],Vars,Xcif,Cif).

%
% add precedence info. using the list of depend/2
%
add_precedence_info([]).
add_precedence_info([depend(X,Y)|Ds]) :-
	sort_precedence:add_precedence(Y,X),
	add_precedence_info(Ds).

%
% a list
%
xcif2cif([],[]).

xcif2cif([X|Xs],[C|Cs]) :-
	xcif2cif(X,C),
	xcif2cif(Xs,Cs).

%
%
%
xcif2cif(Xcif,Cif) :-
	Xcif=restrict_subquery(Q,G,P),
	Q=..[Quantifier|_],
	xcif2cif(P,P_mapped),							%map the predicate
	%
	% choose order of variable to quantify
	%
	Filter_pattern=[	generate(string,_),
				generate(integer,_),
				generate(float,_),
				generate(boolean,_)
			],
	cif_util:match_pattern(G,Filter_pattern,_,G_filtered),

	cif_util:find_var_dependency(G_filtered,Dependency),				%find var dependency in a list of depend/2
	sort_precedence:clean_precedence,
	add_precedence_info(Dependency),
	sort_precedence:sort(Btrees),						%sort vars according to dependency
	sort_precedence:btree2list(Btrees,Vars_list),				%map btrees into list of list
	flatten:flatten(Vars_list,Vars),					%flatten list of list into a list
	make_nested_quantification(Quantifier,[],Vars,G_filtered,P_mapped,Cif).


make_nested_quantification(Quantifier,_,[Var],G,P,Cif) :-
	!,
	Q=..[Quantifier,Var],
	Cif=restrict_subquery(Q,G,P).

make_nested_quantification(Quantifier,Quantified_vars,[Var|Remaining_vars],G,P,Cif) :-
	%
	% we find out the CIF that should not be included in the generator
	%
	cif_util:cif_with_var(Remaining_vars,G,Cif_with_remaining_vars,Remaining_cif),	%find CIF about remaining vars
	This_gen=Remaining_cif,
	make_nested_quantification(Quantifier,Quantified_vars,Remaining_vars,Cif_with_remaining_vars,P,This_p),
	Q=..[Quantifier,Var],
	Cif=restrict_subquery(Q,This_gen,This_p).


%
% CIF is restrict_subquery/3
%
/*
xcif2cif(Vars,Xcif,Cif) :-
	Xcif=restrict_subquery(Q,G,P),
	!,
	Q=..[Quantifier,Vars],
	%
	% filter out generate/2 of the 4 basic types
	%
	Filter_pattern=[	generate(string,_),
				generate(integer,_),
				generate(float,_),
				generate(boolean,_)
			],
	cif_util:match_pattern(G,Filter_pattern,_,G_filtered),
	%
	% now get generate/2 of user-defined types
	%
	Gen2_pattern=[generate(_,_)],
	cif_util:match_pattern(G_filtered,Gen2_pattern,Gen2,Non_gen2),
	Gen5_pattern=[generate(_,_,_,_,_)],
	cif_util:match_pattern(Non_gen2,Gen5_pattern,Gen5,Non_gen),
	xcif2cif(P,P_mapped),						%recursively work on the predicate 1st
	make_nested_cif(Quantifier,Vars,[Gen2,Gen5,Non_gen],Non_gens,P_mapped,Cif).
%	make_nested_cif(Quantifier,Vars,Gens,Non_gens,P_mapped,Cif).
*/

%
% all others
%
xcif2cif(Cif,Cif).

%
% given a list of generate/2, compose a nested constraint
%
make_nested_cif(Quantifier,Vars,[G],Non_gens,P_mapped,Cif) :-
	!,
	Q=..[Quantifier|Vars],
	Cif=restrict_subquery(Q,[G|Non_gens],P_mapped).

make_nested_cif(Quantifier,Vars,[G|Gs],Non_gens,P_mapped,Cif) :-
	G=generate(_,Var),
	Q=..[Quantifier,Var],
	Cif=restrict_subquery(Q,G,Temp_cif),
	%
	%
	%
	delete(Vars,Var,Temp_vars),				%remove the quantified var
	make_nested_cif(Quantifier,Temp_vars,Gs,Non_gens,P_mapped,Temp_cif).
